#include <assert.h>
#include <fstream>
#include "tokenizer.h"

/**
 * class StreamTokenizer
 * 
 * This class is basically a clone of Java's StreamTokenizer (but less 
 * powerful of course). See 
 * 
 *     http://java.sun.com/j2se/1.4/docs/api/java/io/StreamTokenizer.html
 *
 * It takes an input stream and parses it into "tokens", allowing the tokens 
 * to be read one at a time. The parsing process is controlled by a table 
 * and a number of flags that can be set to various states. The stream 
 * tokenizer can recognize identifiers, numbers, quoted strings, and various 
 * comment styles. 
 */

/**
 * Static members have to be initialized in a definition outside 
 * the class declaration. Non-static const members can be initialized 
 * in initialization list of the constructor.
 */
const int StreamTokenizer::MAX_TOKENLEN = MAX_LEN;
const int StreamTokenizer::TT_EOF = 0;
const int StreamTokenizer::TT_EOL = 1; 
const int StreamTokenizer::TT_NUMBER = 2; 
const int StreamTokenizer::TT_WORD = 3;   


#ifdef _MSC_VER
StreamTokenizer::StreamTokenizer(ifstream &file) {
    m_file = file;
    initialize();
}
#endif
/** gcc's std::ios_base::operator = is private. A bug? */

StreamTokenizer::StreamTokenizer(char * filename) {
    m_filename = filename;
    m_file.open(m_filename);
    initialize();
}

void StreamTokenizer::initialize() {
    m_lineno = 0;
    m_eolIsSignificant = false;
    m_slashSlashComments = false;
    m_slashStarComments = false;
    m_pushedBack = false;
    m_parseNumbers = false;
    m_lowerCaseMode = false;
    m_prevNewline = false;

    // Defaults:
    m_commentChar = '#';
    m_quoteChar = '\"';
    m_whitespaceCharLow = 0x00;
    m_whitespaceCharHigh = 0x20;
    m_wordCharLow = 0x21;
    //    m_wordCharHigh = 0xff;
    m_wordCharHigh = 0x7f;
    
    // Ordinary chars are not delt with yet.
    m_ordinaryCharLow = 0;
    m_ordinaryCharHigh = 0;  
}

void StreamTokenizer::resetSyntax() {
    initialize();
    // What else?
}

/**
 * Determines whether or not ends of line are treated as tokens. 
 * If the flag argument is true, this tokenizer treats end of lines 
 * as tokens; the nextToken method returns TT_EOL and also sets the 
 * ttype field to this value when an end of line is read. 
 * 
 * A line is a sequence of characters ending with either a carriage-
 * return character ('\r') or a newline character ('\n'). In addition, 
 * a carriage-return character followed immediately by a newline 
 * character is treated as a single end-of-line token. 
 *
 * If the flag is false, end-of-line characters are treated as white 
 * space and serve only to separate tokens. 
 * 
 * Parameters:
 * flag - true indicates that end-of-line characters are separate tokens; 
 * false indicates that end-of-line characters are white space.
 */
void StreamTokenizer::eolIsSignificant(bool flag) { 
    m_eolIsSignificant = flag; 
} 


/** 
 * Parses the next token from the input stream of this tokenizer. The type 
 * of the next token is returned in the ttype field. Additional information 
 * about the token may be in the nval field or the sval field of this 
 * tokenizer. Typical clients of this class first set up the syntax tables 
 * and then sit in a loop calling nextToken to parse successive tokens until 
 * TT_EOF is returned.
 * 
 * Returns:
 * 
 * the value of the ttype field. 
 */
int StreamTokenizer::nextToken() {
    char c;
    int nchars;
    bool in_comment, in_token, in_quote;

    if(m_pushedBack) {   // Token has been put back, use it
	m_pushedBack = false;
	return ttype;
    }
    if(m_prevNewline) { // Last time we saw a new line. Return it this time.
	m_prevNewline = false;
	//	m_lineno++;

	if(m_eolIsSignificant) { // A new line
	    ttype = TT_EOL;
	    sval[0] = '\0';
	    return ttype;  // Return new line as a token	
	}
    }
    in_comment = false;
    in_token = false;
    in_quote = false;
    nchars = 0;
    
    /** 
     * Editor may use '\r' or '\n' or '\r' followed immediately 
     * by '\n'as EOL. 
     */
    do {
	c = m_file.get();
	if(in_comment) {
	    if(isNewline(c)) {
		in_comment = false;
		m_lineno++;
		if( m_eolIsSignificant) { // A new line
		    ttype = TT_EOL;  // Return new line as a token
		    break;
		}

		continue;
	    }
	    continue;
	}

	/** Note we don't treat = as special. */
//  	if(c == '=' && !in_token) { // First see =, should be treated as a token.
//  	    sval[nchars++] = c;
//  	    ttype = TT_WORD;
//  	    break;
//  	}

//         	if(isWhitespace(c) || c == '=') {
	if(isWhitespace(c)) { /** Note we don't treat = as special. */
	    if(isNewline(c)) // Whitespace can also be new line 
		m_lineno++;

	    if(in_token){  // We are in the middle of a token. This should end it.
		if(c == '=')
		    m_file.unget();  // Put back the =
		
		sval[nchars] = '\0';  // Must terminate the str before calling tokenType.
		ttype = tokenType();
		break;
	    }

	    continue;    // Ignore the leading whitespaces before a token
	}
	
	if(c == m_commentChar) {
	    assert(in_comment == false);
	    in_comment = true;

	    if(in_token){  // A comment should also end a token.
		sval[nchars] = '\0';  // Must terminate the str before calling tokenType.
		ttype = tokenType();
		break;
	    }
	    continue;
	}

	if(isNewline(c)) {
	    m_lineno++;

	    if(in_token){  // A new line should also end a token.
		sval[nchars] = '\0';  // Must terminate the str before calling tokenType.
		ttype = tokenType();
		if( m_eolIsSignificant){  // A new line is counted as a token
		    // m_file.unget();   can't put it back?
		    m_prevNewline = true;
		// Put it back so next time we can return it as token
		}
		break;
	    }

	    if(m_eolIsSignificant) { // A new line
		ttype = TT_EOL;
		return ttype;  // Return new line as a token
	    }
	    else
		continue;
	}
	if (c == EOF ) {
	    if(in_token){  // EOF should also end a token.
		sval[nchars] = '\0';  // Must terminate the str before calling tokenType.
		ttype = tokenType();
		break;
	    }
	    ttype = TT_EOF;
	    break;
	}
	
	// We don't support nested quote
	if(c == m_quoteChar) {
	    if(in_quote)  // end a quote and return a token.
		break;
	    else {
		in_quote = true;
		continue;
	    }
	}
	// Now we are really in the middle of a token.
	in_token = true;
	if(m_lowerCaseMode)
	    sval[nchars++] = tolower(c);
	else
	    sval[nchars++] = c;

    } while (m_file.good());
    
    sval[nchars] = '\0';
    return ttype; 
}

bool StreamTokenizer::isWhitespace(char c) {
    /**
     * Note that '\r' and '\n\ may also be with the range from 
     * m_whitespaceCharLow to m_whitespaceCharHigh. If m_eolIsSignificantis 
     * true, we must treat them as token even if they are within that range.
     */
    if(m_eolIsSignificant && (c == '\r' || c == '\n'))
	return false;

    return ((m_whitespaceCharLow <= c 
	     && c <= m_whitespaceCharHigh) 
	    || (c == '\r' || c == '\n')); 
}

bool StreamTokenizer::isNewline(char c) {
    if(c == '\r') {
	c = m_file.get();   // See if next char is '\n'
	if ( c != '\n' )
	    m_file.unget(); // Put it back if not
	
	return true;  // '\r' followed by '\n' or '\r' along are both newline
    }
    
    if(c == '\n')
	return true;
    return false;
}

/** 
 * Try to distinguish number and word. EOF and EOL are handled by nextToken. 
 * If it sval is a number, its value will be read to nval. Note that we only
 * recognize [-][0x]xxx[.xxx] style numbers. No 10e3 stuff.
*/
int StreamTokenizer::tokenType() {
    char *p = sval;
    int ret = TT_WORD, test;
    bool dot_seen, minus_seen, is_hex;

    dot_seen = minus_seen = is_hex = false;
    if(sval[0] == '0' && sval[1] == 'x')
	is_hex = true;
    
    // We will test if the string is a number. Other wise it is a word.
    while(*p) {
	if(is_hex)
	    test = isxdigit(*p);
	else
	    test = isdigit(*p);
	
	if(!test) {  // Not a digit
	    if(*p == '-') {
		if(minus_seen)  // Can only see - once
		    return ret;
		else 
		    minus_seen = true;
	    }
	    else if(*p == '.') {
		if(dot_seen)  // Can only see . once
		    return ret;
		else 
		    dot_seen = true;
	    }
	    else
		return ret; 
	    // . and - are the only two non-digit chars allowed
	}
	p++;
    }
    ret = TT_NUMBER;
    // Lets copy the number to nval
    nval = atof(sval);
    return ret;
}
