#ifndef __TOKENIZER_H
#define __TOKENIZER_H

#include <fstream>
#include <string>

/**
 * class StreamTokenizer
 * 
 * This class is basically a clone of Java's StreamTokenizer (but less 
 * powerful of course). See 
 * 
 *     http://java.sun.com/j2se/1.4/docs/api/java/io/StreamTokenizer.html
 *
 * It takes an input stream and parses it into "tokens", allowing the tokens 
 * to be read one at a time. The parsing process is controlled by a table 
 * and a number of flags that can be set to various states. The stream 
 * tokenizer can recognize identifiers, numbers, quoted strings, and various 
 * comment styles. 
 */
using namespace std;

#define MAX_LEN 256

class StreamTokenizer {
    // Public fields:
 public:
    static const int MAX_TOKENLEN;
    /**
     * If the current token is a number, this field contains 
     * the value of that number. 
     */
    double nval; 
    /** 
     * If the current token is a word token, this field contains 
     * a string giving the characters of the word token.
     */
    char sval[MAX_LEN]; 
    /** 
     * After a call to the nextToken method, this field contains the 
     * type of the token just read. 
     */
    int ttype;
    /** A constant indicating that the end of the stream has been read. */
    static const int TT_EOF; 
    /** A constant indicating that the end of the line has been read.  */
    static const int TT_EOL; 
    /** A constant indicating that a number token has been read. */
    static const int TT_NUMBER; 
    /** A constant indicating that a word token has been read.  */
    static const int TT_WORD;   

    // Public functions:
 public:
    // Constructor:

#ifdef _MSC_VER
    StreamTokenizer(ifstream &file);
#endif
    /** gcc's std::ios_base::operator = is private. A bug? */

    StreamTokenizer(char * filename);
    ~StreamTokenizer() { m_file.close(); }

    /** Specified that the character argument starts a single-line comment.*/
    void commentChar(int ch) { m_commentChar = ch; } 
    /** Determines whether or not ends of line are treated as tokens. */
    void eolIsSignificant(bool flag);
    /** Return the current line number. */
    int lineno() const { return m_lineno; }
    /** Determines whether or not word token are automatically lowercased. */
    void lowerCaseMode(bool fl) {  m_lowerCaseMode = fl; }
    /** Parses the next token from the input stream of this tokenizer. */
    int nextToken();
    /** 
     * Specifies that the character argument is "ordinary" in this tokenizer.
     * It removes any special significance the character has as a comment character, 
     * word component, string delimiter, white space, or number character. When such 
     * a character is encountered by the parser, the parser treates it as a single-
     * character token and sets ttype field to the character value. 
     */

    // NOT DEALING WITH "ORDINARY" CHARS YET.
    void ordinaryChar(int ch); 
    /** 
     * Specifies that all characters c in the range low <= c <= high are 
     * "ordinary" in this tokenizer. 
     */    
    void ordinaryChars(int low, int hi) {
	m_ordinaryCharLow = low;
	m_ordinaryCharHigh = hi;
    }
    /** Specifies that numbers should be parsed by this tokenizer. */
    void parseNumbers();    // NOT IMPLEMENTED
    /**
     * Causes the next call to the nextToken method of this tokenizer to return 
     * the current value in the ttype field, and not to modify the value in the 
     * nval or sval field. 
     */
    void pushBack() { m_pushedBack = true; }
    /** 
     * Specifies that matching pairs of this character delimit string constants 
     * in this tokenizer. 
     */
    void quoteChar(int ch) { m_quoteChar = ch; } 
    /** 
     * Resets this tokenizer's syntax table so that all characters are "ordinary." 
     * See the ordinaryChar method for more information on a character being ordinary. 
     */
    void resetSyntax();
    /** Determines whether or not the tokenizer recognizes C++-style comments. */     
    void slashSlashComments(bool flag) { m_slashSlashComments = flag; }
    /** Determines whether or not the tokenizer recognizes C-style comments. */
    void slashStarComments(bool flag) { m_slashStarComments = flag; }
    /** Returns the string representation of the current stream token. */
    string toString() { return "StreamTokenizer"; }
    /** 
     * Specifies that all characters c in the range low <= c <= high are white 
     * space characters. 
     */
    void whitespaceChars(int low, int hi) {
	m_whitespaceCharLow = low;
	m_whitespaceCharHigh = hi;
    }
    /** 
     * Specifies that all characters c in the range low <= c <= high are word 
     * constituents. 
     */
    void wordChars(int low, int hi) {
	m_wordCharLow = low;
	m_wordCharHigh = hi;
    }

    // Private fields:
 private:
    ifstream m_file;
    char * m_filename;
    int m_lineno;
    
    // Flags:
    bool m_eolIsSignificant;
    bool m_slashSlashComments;
    bool m_slashStarComments;
    bool m_pushedBack;
    bool m_parseNumbers;
    bool m_lowerCaseMode;

    bool m_prevNewline; 

    char m_commentChar;
    char m_quoteChar;
    char m_whitespaceCharLow;
    char m_whitespaceCharHigh;
    char m_wordCharLow;
    char m_wordCharHigh;
    char m_ordinaryCharLow;
    char m_ordinaryCharHigh;

    // Private functions:
 private:
    void initialize();
    bool isNewline(char c); 
    bool isWhitespace(char c);
    int tokenType();
    
};

#endif
