/** 
 * Turn off the long identifer string warning in debug mood in VC++6.0.
 */

#ifdef _MSC_VER
#ifdef _DEBUG
#pragma warning(disable:4786)
#endif
#endif

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "ucsecconfig.h"


/** 
 * Compare string ignoring case. _stricmp works only on Windows
 * and strcasecmp conforms to BSD4.4.
 */

#ifdef _MSC_VER
#define _STRCMP_NOCASE(s1, s2)  _stricmp(s1, s2)
#else 
#define _STRCMP_NOCASE(s1, s2) strcasecmp(s1, s2)
#endif



/**
 * Static members have to be initialized in a definition outside 
 * the class declaration. Non-static const members can be initialized 
 * in initialization list of the constructor.
 */
const int Value::VT_UNKNOWN = -1;
const int Value::VT_INTEGER = 0;
const int Value::VT_STRING = 1; 
const int Value::VT_BOOLEAN = 2; 
const int Value::VT_DOUBLE = 3;   

/**
 * Note that gcc seems to only allow using classname::varialename to reference
 * static members of a class. With MSVC++6.0, you can do the same or use 
 * classname.varialename. E.g.
 *
 *    (1) StreamTokenizer::MAX_TOKENLEN
 *    (2) StreamTokenizer.MAX_TOKENLEN
 *
 * Both (1) and (2) work with MSVC++6.0 but gcc only allows (1). This file is 
 * using (1) so it can be compiled by both.
 */


Value::Value() {
    vtype = VT_UNKNOWN;
    memset(val.strValue, '\0', StreamTokenizer::MAX_TOKENLEN);
}


/**
 * class Confign. Provides facilities to load and set figuration. The configuration
 * file uses 'variable = value' and '#' comments syntax. 
 */


/**
 * Load configuration information from the specified file. Must be called
 * before any session starts.
 *
 * @param	fileName	the name of the file containing the
 *				configuration to use.
 */
void Config::Load(const char* fileName) {
    char msg[64];
    //    Value value;

    configFile = (char *)fileName;
    StreamTokenizer s(configFile);

    assert(!loaded);
    loaded = true;
	
    s.resetSyntax();
    s.whitespaceChars(0x00, 0x20);
    s.wordChars(0x21, 0xFF);
    s.eolIsSignificant(true);
    s.commentChar('#');
    s.quoteChar('"');
    
    int line = 1;
    
    s.nextToken();
    
    while(true) {
	if(s.ttype == StreamTokenizer::TT_EOF)
	    break;
	
	if(s.ttype == StreamTokenizer::TT_EOL) {
	    line++;
	    s.nextToken();
	    continue;
	}
	
	// First token should be the key
	if(s.ttype != StreamTokenizer::TT_WORD)
	    loadError(line);
	string key = s.sval;
	
	// Next should be =
	if(s.nextToken() != StreamTokenizer::TT_WORD ||
	   strcmp(s.sval, "=") != 0)
	    loadError(line);
	
	// Now the value
	s.nextToken();
	Value value;
	
	if(s.ttype == StreamTokenizer::TT_WORD){
	    if(_STRCMP_NOCASE(s.sval, "true") == 0) {
		value.vtype = Value::VT_BOOLEAN;
		value.val.boolValue = true;
	    }
	    else if(_STRCMP_NOCASE(s.sval, "false") == 0){
		value.vtype = Value::VT_BOOLEAN;
		value.val.boolValue = false;
	    }
	    else {// A string
		value.vtype = Value::VT_STRING;
		strcpy(value.val.strValue, s.sval);
	    }
	}
	else if(s.ttype == StreamTokenizer::TT_NUMBER) {
	    value.vtype = Value::VT_DOUBLE;
	    value.val.doubleValue = s.nval;
	}
	else  // Error
	    loadError(line);
	
	// ignore everything after first string
	while (s.nextToken() != StreamTokenizer::TT_EOL &&
	       s.ttype != StreamTokenizer::TT_EOF);
	
	// See if we have this key already
	CfgMap::iterator theIterator =  config.find(key);
	if(theIterator != config.end()) {
	    sprintf(msg, "duplicate key %s.\n", key.c_str());
	    configError(msg);
	}
	
	config.insert(CfgMap::value_type(key, value));
	line++;
    }
}

/**
 * Get the value of a key in configuration file.
 *
 * @param	key	the key to look up.
 * @return	the value of the specified key, or NULL if it is not
 *		present.
 */
char * Config::GetString(const string key) {
    char msg[64];
    CfgMap::iterator theIterator =  config.find(key);
    if(theIterator != config.end()) {
	Value sv = (*theIterator).second;
	if(sv.vtype != Value::VT_STRING) {
	    sprintf(msg, "%s should be string.\n", key.data());
	    configError(msg);
	}
	return (*theIterator).second.val.strValue;
    }
    else {
	sprintf(msg, "no such key: %s.\n", key.data());
	configError(msg);
	return "";
    }
}

/**
 * Get the value of an integer key in configuration file.
 *
 * @param	key	the key to look up.
 * @return	the value of the specified key.
 */
int Config::GetInteger(const string key) {
    char msg[64];
    CfgMap::iterator theIterator =  config.find(key);
    if(theIterator != config.end()) {
	Value sv = (*theIterator).second;
	if(sv.vtype != Value::VT_INTEGER) {
	    sprintf(msg, "%s should be integer.\n", key.data());
	    configError(msg);
	}
	return sv.val.intValue;
    }
    else {
	sprintf(msg, "no such key: %s.\n", key.data());
	configError(msg);
	return 0;
    }
}

/**
 * Get the value of a double key in configuration file.
 *
 * @param	key	the key to look up.
 * @return	the value of the specified key.
 */
double Config::GetDouble(const string key) {
    char msg[64];
    CfgMap::iterator theIterator =  config.find(key);
    if(theIterator != config.end() ) {
	Value sv = (*theIterator).second;
	if(sv.vtype != Value::VT_DOUBLE) {
	    sprintf(msg, "%s should be double.\n", key.data());
	    configError(msg);
	}
	return sv.val.doubleValue;
    }
    else {
	sprintf(msg, "no such key: %s.\n", key.data());
	configError(msg);
	return .0;
    }
}

/**
 * Get the value of a boolean key in configuration file.
 *
 * @param	key	the key to look up.
 * @return	the value of the specified key.
 */
bool Config::GetBoolean(const string key) {
    char msg[64];
    CfgMap::iterator theIterator =  config.find(key);
    if(theIterator != config.end() ) {
	Value sv = (*theIterator).second;
	if(sv.vtype != Value::VT_BOOLEAN) {
	    sprintf(msg, "%s should be boolean.\n", key.data());
	    configError(msg);
	}
	return sv.val.boolValue;
    }
    else {
	sprintf(msg, "no such key: %s.\n", key.data());
	configError(msg);
	return false;
    }
}

