#ifndef __UCSEC_CONFIG_H
#define __UCSEC_CONFIG_H
// Can't call this file config.h because Crypto++ has one with the same name.

#include <string>
#include <map>
// STL headers
#include "tokenizer.h"

union ValueField {
    int intValue;
    char strValue[MAX_LEN];
    double doubleValue;
    bool boolValue;
};

class Value {
 public:
    static const int VT_UNKNOWN;
    static const int VT_INTEGER;
    static const int VT_STRING;
    static const int VT_BOOLEAN;
    static const int VT_DOUBLE;

    ValueField val;
    int vtype;

    Value();
};


using namespace std;
typedef map<string, Value, less<string> > CfgMap;
/**
 * STL map is not actually hash map. It is a list sorted by key. But it works here.
 * Insertion, look up and deletion take log(n) time.
 */


class CfgException : public exception {
 public:
    explicit CfgException(const string &s) : m_what(s) {}
    virtual ~CfgException() throw() {}
    const char *what() const throw() {return (m_what.c_str());}
    const string &GetWhat() const {return m_what;}
    void SetWhat(const std::string &s) {m_what = s;}

 private:
    string m_what;
};


/**
 * class Config. Provides facilities to load and set figuration. The configuration
 * file uses 'variable = value' and '#' comments syntax. 
 */
class Config {
 public:
    Config() { loaded = false; configFile = NULL; }

    /**
     * Load configuration information from the specified file. Must be called
     * before any session starts.
     *
     * @param	fileName	the name of the file containing the
     *				configuration to use.
     */
    void Load(const char* fileName);

    /**
     * Get the value of a key in configuration file.
     *
     * @param	key	the key to look up.
     * @return	the value of the specified key, or NULL if it is not
     *		present.
     */
    char * GetString(const string key);

    /**
     * Get the value of an integer key in configuration file.
     *
     * @param	key	the key to look up.
     * @return	the value of the specified key.
     */
    int GetInteger(const string key);

    /**
     * Get the value of a double key in configuration file.
     *
     * @param	key	the key to look up.
     * @return	the value of the specified key.
     */
    double GetDouble(const string key);

    /**
     * Get the value of a boolean key in configuration file.
     *
     * @param	key	the key to look up.
     * @return	the value of the specified key.
     */
    bool GetBoolean(const string key);
 private:
    bool loaded;
    char * configFile;
    CfgMap config;

    void loadError(int line) {
	fprintf(stderr, "Error in %s line %d\n", configFile, line);
	exit(1);
    }

    void configError(const char *message) {
	//	fprintf(stderr, "\nError in %s: %s\n", configFile, message);
	throw CfgException(message);
	//exit(1);
    }

};

#endif
